require 'VMParser'

class MiniVMParser < VMParser

  def initialize(file)
	super
	@ops = define_opcodes
	@directives = define_directives
  end

  def define_opcodes
	ops = VMOpcodes.new
	ops.add("PUSH","r1",nil,"\x30")
	ops.add("PUSH","r2",nil,"\x31")
	ops.add("PUSH","r3",nil,"\x32")
	ops.add("PUSH","r4",nil,"\x33")
	ops.add("PUSH","eip",nil,"\x34")
	ops.add("PUSH","esp",nil,"\x35")
	ops.add("PUSH","flags",nil,"\x36")
	ops.add("POP","r1",nil,"\x38")
	ops.add("POP","r2",nil,"\x39")
	ops.add("POP","r3",nil,"\x3a")
	ops.add("POP","r4",nil,"\x3b")
	ops.add("POP","eip",nil,"\x3c")
	ops.add("POP","esp",nil,"\x3d")
	ops.add("POP","flags",nil,"\x3e")
	ops.add("CMP","r1","r2","\xd0")
	ops.add("CMP","r1",:value,"\xd8")
	ops.add("CMP","r2",:value,"\xd9")
	ops.add("CMP","r3",:value,"\xda")
	ops.add("CMP","r4",:value,"\xdb")
	ops.add("CMP","eip",:value,"\xdc")
	ops.add("CMP","esp",:value,"\xdd")
	ops.add("CMP","flags",:value,"\xde")
	ops.add("DEC","r1",nil,"\xe0")
	ops.add("DEC","r2",nil,"\xe1")
	ops.add("DEC","r3",nil,"\xe2")
	ops.add("DEC","r4",nil,"\xe3")
	ops.add("DEC","eip",nil,"\xe4")
	ops.add("DEC","esp",nil,"\xe5")
	ops.add("INC","r1",nil,"\xe8")
	ops.add("INC","r2",nil,"\xe9")
	ops.add("INC","r3",nil,"\xea")
	ops.add("INC","r4",nil,"\xeb")
	ops.add("INC","eip",nil,"\xec")
	ops.add("INC","esp",nil,"\xed")
	ops.add("XOR","r1","r1","\x00")
	ops.add("XOR","r1","r2","\x01")
	ops.add("XOR","r1","r3","\x02")
	ops.add("XOR","r1","r4","\x03")
	ops.add("XOR","r1","eip","\x04")
	ops.add("XOR","r1","esp","\x05")
	ops.add("JMP",:value,nil,"\x10")
	ops.add("DJMP",:value,nil,"\x11")
	ops.add("JE",:value,nil,"\x20")
	ops.add("JL",:value,nil,"\x21")
	ops.add("JG",:value,nil,"\x22")
	ops.add("AND","r1",:value,"\x40")
	ops.add("OR","r1",:value,"\x41")
	ops.add("MOV","r1","[r1]","\x42")
	ops.add("EXIT",nil,nil,"\x43")
	ops.add("CALL","r1",nil,"\x50")
	ops.add("CALL",:value,nil,"\x51")
	ops.add("MOV","r1","r2","\xb0")
	ops.add("MOV","r1","r3","\xb1")
	ops.add("MOV","r1","r4","\xb2")
	ops.add("MOV","r1","eip","\xb3")
	ops.add("MOV","r1","esp","\xb4")
	ops.add("MOV","r1","flags","\xb5")
	ops.add("MOV","[r1]",:value,"\xb8")
	ops.add("MOV","[r2]",:value,"\xb9")
	ops.add("MOV","[r3]",:value,"\xba")
	ops.add("MOV","[r4]",:value,"\xbb")
	ops.add("MOV","[eip]",:value,"\xbc")
	ops.add("MOV","[esp]",:value,"\xbd")
	ops.add("MOV","[flags]",:value,"\xbe")
	ops.add("MOV","r1",:value,"\xc8")
	ops.add("MOV","r2",:value,"\xc9")
	ops.add("MOV","r3",:value,"\xca")
	ops.add("MOV","r4",:value,"\xcb")
	ops.add("MOV","eip",:value,"\xcc")
	ops.add("MOV","esp",:value,"\xcd")
	ops.add("MOV","flags",:value,"\xce")
	ops.add("ADD","r1","r1","\x60")
	ops.add("ADD","r1","r2","\x61")
	ops.add("ADD","r1","r3","\x62")
	ops.add("ADD","r1","r4","\x63")
	ops.add("ADD","r1","eip","\x64")
	ops.add("ADD","r1","esp","\x65")
	ops.add("SUB","r1","r1","\x68")
	ops.add("SUB","r1","r2","\x69")
	ops.add("SUB","r1","r3","\x6a")
	ops.add("SUB","r1","r4","\x6b")
	ops.add("SUB","r1","eip","\x6c")
	ops.add("SUB","r1","esp","\x6d")
	return ops
  end

  def define_directives
	@directives = VMDirectives.new
	@directives.add("db");
	@directives.add("dbx");
	@directives
  end

  def get_registers
	registers = Array.new
	registers.push("r1")
	registers.push("r2")
	registers.push("r3")
	registers.push("r4")
	registers.push("flags")
	registers.push("eip")
	registers.push("esp")
	return registers
  end

  def check_valid(op)
	@ops.opcodes.each do |opc|
		return true if opc.cmd == op
	end
	return false
  end

  def get_arg_min(op)
	min = 2
	@ops.opcodes.each do |opc|
		if opc.cmd == op
			if not opc.arg1
				min = 0
			elsif not opc.arg2 and min > 0
				min = 1
			end
		end
	end
	min
  end

  def get_arg_max(op)
	max = 0
	@ops.opcodes.each do |opc|
		if opc.cmd == op
			if opc.arg2
				max = 2
			elsif opc.arg1 and max < 2
				max = 1
			end
		end
	end
	max
  end

  def get_operand_size(tok)
	size = 0
	@ops.opcodes.each do |opc|
		if opc.cmd == tok.operand
			size = opc.bytes.length
			if tok.arg1.is_a? VMValue or tok.arg2.is_a? VMValue or
			   tok.arg1.is_a? VMLabel or tok.arg2.is_a? VMLabel
				size += 4
			end
			return size
		end
	end
	size
  end

  # Ok we cheat here... :)
  def get_directive_size(tok)
	process_directive(tok, nil).size
  end

  def get_label_pos(label, tokens)
	@tokens.each do |tok|
		return tok.position.to_s if label == tok.label
	end
	throw :couldntFindLabel
  end

  # Splits up strings in this syntax: 'my, string', 10, 0
  # to ["my, string"], [10], [0]
  def string2chunks(str)
	return nil if not str
	instr = false
	notstr = true
	chunk = String.new
	chunks = Array.new
	str.each_byte do |byte|
		if byte == 39  # If ' found toggle instr mode
			if instr
				instr = false
			else
				instr = true 
				notstr = false
			end
		# If a , is found and not in a string and we have a chunk...
		elsif byte == 44 and not instr and chunk.size > 0
			if notstr   # If chunk is not a stirng convert to int
				chunks.push(chunk.to_i)
			else
				chunks.push(chunk)
			end
			notstr = true
			chunk = String.new
		# Is this a space that is not in a string...ignore
		elsif (byte == 32 or byte == 9) and not instr
			# Do nothing
		else
			if instr
				chunk += byte.chr
			elsif byte > 47 and byte < 58
				chunk += byte.chr
			else
				throw :invalidStringFormat
			end
		end
	end
	# Still one more chunk left....
	if chunk.size > 0
		if notstr   # If chunk is not a stirng convert to int
			chunks.push(chunk.to_i)
		else
			chunks.push(chunk)
		end
	end
	chunks
  end

  # Given a string convert it to a binary string
  # line syntax:  'my string',10,13,0
  def make_binary_string(line)
      return nil if not line
	chunks = string2chunks(line)
	bytecode = String.new
	chunks.each do |chunk|
		bytecode << chunk
	end
	bytecode
  end

  def make_xor_binary_string(line)
	return nil if not line
	chunks = string2chunks(line)
	xorkey = chunks.shift
	throw :xorKeyNotFixnum if not xorkey.is_a? Fixnum
	bytecode = String.new
	encbytecode = String.new
	chunks.each do |chunk|
		bytecode << chunk
	end
	bytecode.each_byte do |byte|
		byte = byte ^ xorkey
		encbytecode << byte
	end
	encbytecode
  end

  def process_directive(tok, tokens)
	return nil if not tok.directive
	if(tok.directive.cmd == "db") 
		return make_binary_string(tok.directive.line)
	elsif(tok.directive.cmd == "dbx") 
		return make_xor_binary_string(tok.directive.line)
	else
		puts "Warning: Possibily unsupported directive on line: #{tok.line}  #{tok.directive.cmd}"
	end
	return nil
  end
  
  def get_bytecode(tok, tokens)
	@ops.opcodes.each do |opc|
		begin
			tok.arg1 = VMValue.new(get_label_pos(tok.arg1, tokens)) if tok.arg1.is_a? VMLabel
			tok.arg2 = VMValue.new(get_label_pos(tok.arg2, tokens)) if tok.arg2.is_a? VMLabel
		rescue
			puts "Error on line: " + tok.line.to_s
			puts $!
                        exit(5)
		end
		if opc.match(tok)
			matchfound = true
			addr = String.new
			addr += opc.bytes
			if tok.arg1.is_a? VMValue
				i = tok.arg1.to_i
				addr += "%c%c%c%c" % [i, i >> 8, i >> 16, i >> 32]
			end
			if tok.arg2.is_a? VMValue
				i = tok.arg2.to_i
				addr += "%c%c%c%c" % [i, i >> 8, i >> 16, i >> 32]
			end
			return addr
		end
	end
	return nil
  end
end
