require 'VMOpcodes'
require 'VMDirectives'

class VMapp < String

end

class CodeLine
   attr_reader :line, :number
   attr_writer :line, :number
   def initialize(line)
	@line = line
	@number = Integer
   end
end

class VMValue < String

end

class VMRegister < String

end

class VMLabel < String

end

class VMToken
  attr_writer :operand, :arg1, :arg2, :label, :line, :size, :position, :directive
  attr_reader :operand, :arg1, :arg2, :label, :line, :size, :position, :directive
  def initialize(code)
	@operand = nil
	@arg1 =  nil
	@arg2 = nil
	@label = nil
	@line = nil
	@size = 0
	@position = 0
	@directive = nil;
	if code.line=~/^(\w+):/ 
		@label = VMLabel.new($1)
		@line = code.number
	elsif code.line=~/^\.(\w+)\s+(.+)/
		@directive = VMDirective.new($1, $2)
		@line = code.number
	else 
		line = code.line.dup
		if line=~/^(\w+)/
			@operand = $1;
			line.sub!(/^(\w+)/,"")
			if line=~/(\S+)\s*,\s*(\S+)/
				@arg1 = $1
				@arg2 = $2
			elsif line=~/(\S+)/
				@arg1 = $1
			end
		end	
		@line = code.number
	end
  end

  def argcount
	count = 0
	count += 1 if arg1
	count += 1 if arg2
	count
  end

  def to_s
	if @label
		@label + ": @" + @position.to_s + " [" + @size.to_s + "]"
	elsif @directive
		"." + @directive.cmd + " " + @directive.line + " @" + @position.to_s + " [" + @size.to_s + "]"
	elsif @operand
		str = String.new
		str += @operand
		str += " " + @arg1 if @arg1
		str += ", " + @arg2 if @arg2
		str += " @" + @position.to_s + " [" + @size.to_s + "]"
	else
		""
	end
  end
end

class VMParser
  attr_reader :comment, :VMapp, :registers, :verbose, :style
  attr_writer :comment, :VMapp, :registers, :verbose, :style
  def initialize(file)
	src = File.open(file)
	@lines = Array.new
	@tokens = Array.new
	@ops = VMOpcodes.new
	@directives = VMDirectives.new
	newlines = src.readlines()
	newlines.each_index do |idx|
		line = CodeLine.new(newlines[idx].chomp!)
		line.number = idx + 1
		@lines.push(line)
	end
	@comment = ";"
	@registers = nil
	@VMapp = nil
	@curpos = 0
	@registers = nil
	@verbose = false
	@style = "bin"
	@testvm = String.new
  end

  # Removes comments based on @comment character
  def stripComments()
	newlines = Array.new
	@lines.each do |code|
		code.line.gsub!(/#{@comment}.*/,"");
		newlines.push(code) if code.line.length > 0
	end
	@lines = newlines
  end

  # Removes spaces
  def stripSpaces()
	newlines = Array.new
	@lines.each do |code|
		code.line.strip!
		code.line.gsub!(/\s{2,}/," ");
		newlines.push(code) if code.line.length > 0
	end
	@lines = newlines
  end

  # Convert the code lines into tokens
  def tokenize()
	@lines.each do |code|
		tok = VMToken.new(code)		
		@tokens.push(tok)
	end
  end

  # Fetches an string array of all operands of the type passed
  # Example return all PUSH operands the Compiler supports
  def fetchOperands(op)
	matchingOps = Array.new
	@ops.opcodes.each do |vmop|
		if vmop.cmd == op
			matchingOps.push(vmop.to_s)
		end
	end
	matchingOps
  end

################################
  # Should overide these

  # Checks for a valid operand
  def check_valid(op)
	return true
  end

  # Minimum arguments for a given operand
  def get_arg_min(op)
	return 0
  end

  # Max arguments for a given operand
  def get_arg_max(op)
	return 2
  end

  # Returns an array of supported registers
  # Mainly used for error handling
  def get_registers
	registers = Array.new
	return registers
  end

  # Compiler directives are a special case therefore have their own functions
  # This allows to extend an existing compiler with just special directive
  # extensions
  def get_directive_size(tok)
	return 0
  end

  # Gets the size based on the tokens structure
  # This can be wrong but gives the compiler a good idea where things
  # are in the binary
  def get_operand_size(tok)
	return 0
  end

  # Converts the token to the bytecode
  def get_bytecode(tok, tokens)
	return "\x0" * tok.size
  end

  def verify_directive(tok)
	found = false;
	return true if not tok.directive
	@directives.dirs.each do |dir|
		found = true if dir.cmd == tok.directive.cmd
	end
	found
  end

  # If you want to support directives override this
  # process_directives can also return bytecode if you want
  def process_directive(tok, tokens)
	return nil
  end

##################################

  # Determines if the argument type is a register, value or a label
  def define_arg_type(arg)
	newarg = nil
	arg.strip!
	barearg = arg.dup
	barearg.gsub!(/[\[\]]/,"")
	@registers.each do |reg|
		newarg = VMRegister.new(arg) if barearg == reg
	end
	if not newarg
		value = nil
		if arg=~/^0x([0-9A-F]+)h*$/i
			value = VMValue.new($1.hex.to_s)
		elsif arg=~/^([10]+)+b$/i   # TODO: Finish
			value = VMValue.new($1)
		elsif arg=~/^(\d+)$/i
			value = VMValue.new($1)
		else
			value = VMLabel.new(arg)
		end
		newarg = value if value
	end
	newarg
  end

  # Define the argument types for a token
  def define_arg_types(tok)
	return true if tok.argcount == 0
	@registers = get_registers
	if tok.arg1
		newarg = define_arg_type(tok.arg1)
		return false if not newarg
		tok.arg1 = newarg
	end
	if tok.arg2
		newarg = define_arg_type(tok.arg2)
		return false if not newarg
		tok.arg2 = newarg
	end
	true
  end

  # checks to see if token uses a label and if so verifies it
  def verify_label(tok)
	found = false
	return true if not tok.arg1.is_a? VMLabel and not tok.arg1.is_a? VMLabel
	@tokens.each do |tokl|
		if tok.arg1.is_a? VMLabel
			found = true if tokl.label == tok.arg1
		end
		if tok.arg2.is_a? VMLabel
			found = true if tokl.label == tok.arg2
		end
	end
	return found
  end


  # Validate the code
  def validate()
	@tokens.each do |tok|
		if tok.operand
		   begin
			throw :badOperand if not check_valid(tok.operand)
			min = get_arg_min(tok.operand)
			max = get_arg_max(tok.operand)
			throw :badArgCount if tok.argcount < min or tok.argcount > max
			throw :badArgType if not define_arg_types(tok)
			throw :badLabel if not verify_label(tok)
		   rescue
			puts "Error on line: " + tok.line.to_s
			puts $!
			exit(5)
		   end
		elsif tok.directive
		   begin
			throw :unsupportedDirective if not verify_directive(tok)
		   rescue
			puts "Error on line: " + tok.line.to_s
			puts $!
			exit(15)
		   end
		end
	end
  end

  # Pre allocate where in memory a token will be in the end binary
  # This defines tok.position which is important for labels and whatnot
  def alloc_memory()
	@tokens.each do |tok|
		if tok.directive
			tok.size = self.get_directive_size(tok)
		else
			tok.size = self.get_operand_size(tok)
		end
		tok.position = @curpos
		@curpos += tok.size
	end
  end

  # This prepares all the tokens and validates the code
  def parse()
	puts "VMParser: Parsing code..." if @verbose
	self.stripComments()
	self.stripSpaces()
	self.tokenize()
	puts "VMParser: Validating code..." if @verbose
	self.validate()
  end

  # Main routine.  After the code is handled it write the binary.
  def compile()
	self.parse
	puts "VMParser: Compiling..." if @verbose
	self.alloc_memory
	@VMapp = VMapp.new("\0" * @curpos)
	@tokens.each do |tok|
		if tok.directive
			bytecode = process_directive(tok, @tokens)
		else
			bytecode = get_bytecode(tok, @tokens)
		end
		if bytecode == nil and tok.operand
			puts "Error on line: " + tok.line.to_s
			puts "Unsupported use of " + tok.operand.to_s
			puts "Possible Operand Methods:"
			fetchOperands(tok.operand).each do |suggestion|
				puts "  [ " + suggestion + " ]"
			end
			exit(6)
		end
		if @verbose
			puts "VMParser: " + tok.to_s
			puts "VMParser: " + bytecode.unpack("H*").inspect if bytecode
		end
		@VMapp[tok.position,bytecode.length] = bytecode if bytecode
	end
  end

  # Dumps the code how the VMParser sees it
  def dumpsrc()
	@lines.each do |code|
		puts code.line
	end
  end

  # Dumps C styled array
  def dumpc()
	output = "unsigned char vmcode[] = {\n\t"
	first = true
	count = 1
	total = 1
	@VMapp.each_byte do |byte|
		output += "0x%02x" % byte.to_i
		output += ", " if not total == @VMapp.length
		if count == 10
			output += "\n\t" if not total == @VMapp.length
			count = 0
		end
		count += 1
		total += 1
		first = false if first
	end
	output += "\n};\n"
	output += "unsigned int vmcode_len = #{@VMapp.length};\n"
	output
  end

  # Dumps in MASM format
  def dumpmasm()
	output = String.new
	cnt = 0
	labelcnt = 1
	newlabel = true
	@VMapp.each_byte do |byte|
		output += "vmcode#{labelcnt}\t\tdb " if newlabel
		output += ", " if newlabel == false
		if byte >= 160
			output += "0%02xh" % byte.to_i
		else
			output += "%02xh" % byte.to_i
		end
		cnt += 1
		if cnt == 10
			output += "\n"
			cnt = 0
			newlabel = true
			labelcnt += 1
		else
			newlabel = false
		end
	end
	output
  end

  # Handles different output formats
  def dumpvm()
	if @style =~/^bin/i
		return @VMapp
	end
	if @style =~/^c/i
		return dumpc
	end
	if @style =~/^masm/i
		return dumpmasm
	end
	throw :unkownOutputStyle
  end

end
